/*
 * This is the implementation file for the capability layer
 * of Efficient Capability-Based Messaging system
 *
 * It defines all the functions nedded to manage capabilities
 */

#include <linux/linkage.h>  /* For asmlinkage */
#include <linux/smp_lock.h> /* For indivisibility */
#include <asm/semaphore.h>  /* For indivisibility */
#include <asm/atomic.h>     /* For indivisibility */
#include <linux/malloc.h>   /* For kmalloc */
#include <linux/mm.h>       /* For mm structs */
#include <linux/mman.h>     /* For mm structs */
#include <linux/pagemap.h>  /* For wait_on_page */
#include <asm/page.h>       /* For page related constants */
#include <asm/pgtable.h>    /* For pte related functions */
#include <linux/sched.h>    /* For capability_table struct */
#include <linux/unistd.h>   /* For errno */
#include <linux/errno.h>    /* For errno codes */

#include <linux/ecbm/cap_types.h>
#include <linux/ecbm/cap.h>

extern unsigned long vm_flags(unsigned long prot, unsigned long flags);

/*
 * Get a capability from a virtual memory area 
 * of contigous logical addresses
 * @param void* buffer - the initial address of the area
 * @param int size - the dimension of the area in bytes
 * @param prot rights - the access rights to be granted
 *                      (they will be anded with owner rights)
 * @return a new capability struct describing the message
 */
capability* ecbm_create_capability(void* buffer, int size, prot_t rights)
{
 logical_address_t logical;

 pgd_t* pgd;
 pmd_t* pmd;
 pte_t* pte;
 struct vm_area_struct* vma;
 int write_access;

 struct page* page;
 capability* cap;

 int pages;
 int count;

 /* Get logical address space information */
 logical=(logical_address_t)buffer;
 vma=find_vma_intersection(current->mm,logical,logical+size);
 if(!vma) 
 {
  errno=-EINVAL;
  return NULL;
 }
 write_access=((vma->vm_flags&VM_WRITE)!=0);

 /* Create a new capability struct */
 cap=(capability*)kmalloc(sizeof(capability),GFP_KERNEL);
 cap->offset_start=logical&(~PAGE_MASK);
 pages=(size+cap->offset_start+PAGE_SIZE-1)>>PAGE_SHIFT; 
 cap->addresses=(physical_address_t*)kmalloc(pages*sizeof(physical_address_t),GFP_KERNEL);
 cap->size=pages;
 cap->rights=rights&vma->vm_flags;

 /* Scan pages pages starting at logical and fill in cap */
 down(&current->mm->mmap_sem); /* CRITICAL SECTION: Lock structure, please */
 lock_kernel();
 for(count=0; count<pages; ++count)
 {
  pgd=pgd_offset(current->mm,logical);
  pmd=pmd_offset(pgd,logical);
  pte=pte_offset(pmd,logical);

  if(!pte_present(*pte))
   if(handle_mm_fault(current,vma,logical,write_access)<0)
   {
    errno=-EAGAIN;
    return NULL;
   }

  /* Increment count and lock page in memory*/
  page=&mem_map[MAP_NR(pte_page(*pte))];
  atomic_inc(&page->count);
  set_bit(PG_locked,&page->flags);

  cap->addresses[count]=pte_page(*pte);

  logical+=PAGE_SIZE;
 }
 unlock_kernel();
 up(&current->mm->mmap_sem); /* CRITICAL SECTION: Unlock structure, please */

 return cap;
}

/*
 * Acquire a capability
 * @param capability* cap - the capability to be acquired
 * @param prot_t rights - the access rights to be granted
 *                        (they will be anded with the rights in cap)
 * @return a new virtual memory area of contigous logical addresses 
 *         corresponding to cap
 */
void* ecbm_acquire_capability(capability* cap, prot_t rights)
{ 
 logical_address_t logical;
 logical_address_t ret;

 pgd_t* pgd;
 pmd_t* pmd;
 pte_t* pte;
 pgprot_t new_rights;

 int count;
 struct capability_table* caps;

 /* Allocate a new vm_area_struct and fill in process page table from cap */
 down(&current->mm->mmap_sem); /* CRITICAL SECTION: Lock structure, please */
 lock_kernel();
 ret=do_mmap(NULL,0,cap->size*PAGE_SIZE,rights&cap->rights,MAP_PRIVATE,0);
 logical=ret;
 new_rights=protection_map[(vm_flags(rights&cap->rights,MAP_PRIVATE)|current->mm->def_flags)&0x0f];

 for(count=0; count<cap->size; ++count)
 {
  pgd=pgd_offset(current->mm,logical);
  pmd=pmd_alloc(pgd,logical);
  if(!pmd)
  {
   errno=-ENOMEM;
   return NULL;
  }
  pte=pte_alloc(pmd,logical);
  if(!pte)
  {
   errno=ENOMEM;
   return NULL;
  }
  set_pte(pte,mk_pte(cap->addresses[count],new_rights)); 

  logical+=PAGE_SIZE;
 }
 unlock_kernel();
 up(&current->mm->mmap_sem); /* CRITICAL SECTION: Unlock structure, please */

 /* Adjust address */
 ret+=cap->offset_start;

 /* Create a new capability_table struct */
 caps=(struct capability_table*) kmalloc(sizeof(struct capability_table),GFP_KERNEL); 
 caps->logical=ret;
 caps->cap=cap;

 /* Insert capability in per-process capability table */
 ecbm_insert_capability(caps);
 
 return (void*)(ret);
}

/*
 * Insert a new capability into the pre-process capability table
 * @param struct capability_table* caps - the capability to be inserted
 * @return null
 */
void ecbm_insert_capability(struct capability_table* caps)
{
 struct capability_table* parent;
 struct capability_table* cur;

 /* First capability? */
 if(!current->caps)
 { 
  current->caps=caps;
  caps->next=NULL;
  return;
 }

 /* ORDERED LIST */
 parent=NULL;
 cur=current->caps;
 while(cur && cur->logical<caps->logical)
 {
  parent=cur;
  cur=cur->next;
 }

 if(parent) 
 {
  parent->next=caps;
  caps->next=cur;
 }
 else
 {
  caps->next=current->caps;
  current->caps=caps;
 }
 return;
}

/*
 * Release a virtual memory area previously acquired
 * @param capability* cap - the capability corresponding 
 *                          to the area being released
 * @return TRUE on success, set errno on error
 * @note THE CAPABILITY CAP MUST BE NON NULL!!!
 */
int ecbm_release_capability(capability* cap)
{
 struct page* page;

 int count;

 /* Scan size pages and undo create_capability */
 for(count=0; count<cap->size; ++count)
 {
  /* Decrement count and unlock page in memory*/
  page=&mem_map[MAP_NR(cap->addresses[count])];
  atomic_dec(&page->count);
  clear_bit(PG_locked,&page->flags);
 }

 return TRUE;
}

/*
 * Release a virtual memory area previously acquired
 * @param void* buffer - the initial address of the area
 * @return TRUE on success, set errno on error
 */
asmlinkage int ecbm_release_area(void* buffer)
{
 capability* cap;

 /* Do we really hold this capability? */
 if((cap=ecbm_unlink_capability((logical_address_t)buffer)))
 {
  /* Remove logical mapping */
  down(&current->mm->mmap_sem); /* CRITICAL SECTION: Lock structure, please */
  lock_kernel();
  do_munmap((logical_address_t)buffer,cap->size*PAGE_SIZE);
  unlock_kernel();
  up(&current->mm->mmap_sem); /* CRITICAL SECTION: Unlock structure, please */

  /* Undo create_capability */
  return ecbm_release_capability(cap);
 }

 return -EINVAL;
}

/*
 * Unlink a capability from the pre-process capability table
 * @param logical_address_t logical - the logical address to search for
 * @return capability* the capability corresponding the logical, if any
 */
capability* ecbm_unlink_capability(logical_address_t logical)
{
 struct capability_table* parent;
 struct capability_table* cur;

 capability* ret;

 /* ORDERED LIST */
 parent=NULL;
 cur=current->caps;
 while(cur && cur->logical<logical)
 {
  parent=cur;
  cur=cur->next;
 }
 
 /* Found? */
 if(cur && cur->logical==logical)
 {
  if(parent) 
   parent->next=cur->next;
  else
   current->chs=NULL;

  /* Free some kernel space */
  ret=cur->cap;
  kfree(cur);
  return ret;
 }

 return NULL;
}



