/*
 * This is the implementation file for the communication layer
 * of Efficient Capability-Based Messaging system
 *
 * It defines all the functions nedded to manage communication channels
 */

#include <linux/linkage.h> /* For asmlinkage */
#include <asm/spinlock.h>  /* For indivisibility */
#include <linux/malloc.h>  /* For kmalloc */
#include <linux/mm.h>      /* For mm structs */
#include <linux/wait.h>    /* For wait_queue struct */
#include <linux/sched.h>   /* For channel_table struct */
#include <linux/unistd.h>  /* For errno */
#include <linux/errno.h>   /* For errno codes */

#include <linux/ecbm/cap_types.h>
#include <linux/ecbm/cap.h>
#include <linux/ecbm/comm_types.h>
#include <linux/ecbm/comm.h>

spinlock_t magic_number_spin=SPIN_LOCK_UNLOCKED;
magic_number_t magic_number;
int count;

spinlock_t comm_structs_spin=SPIN_LOCK_UNLOCKED;
channel* list_head;
channel* avl_root;

/*
 * Reserve a new unique channel id
 * @param null
 * @return a new unique channel id
 */
asmlinkage magic_number_t ecbm_get_magic_number()
{
 int magic;
 
 spin_lock(&magic_number_spin);
 magic=magic_number;
 ++magic_number;
 spin_unlock(&magic_number_spin);

 return magic; 
}

/*
 * Install a new channel into kernel structures
 * @param magic_number_t magic - the progressive id of the channel
 * @param int asynchrony - the asynchrony degree of communication
 * @return the address of the channel within the per-process channel table
 */
asmlinkage id_t ecbm_install_channel(magic_number_t magic, int asynchrony)
{
 channel* ch;
 channel* found;
 struct channel_table* chs;

 if(current->chs && current->chs->free_id>=ECBM_COMM_PP_MAX)
  return -EMFILE;

 ch=(channel*)kmalloc(sizeof(channel),GFP_KERNEL);

 /* Search and create channel */
 spin_lock(&comm_structs_spin); /* CRITICAL SECTION: Lock structure, please */
 if(!(found=ecbm_search_channel(magic)))
 {
  ch->magic=magic;
  ch->filled=0;
  init_waitqueue(&ch->acquire);
  ch->spin=SPIN_LOCK_UNLOCKED;
  init_waitqueue(&ch->full);
  init_waitqueue(&ch->empty);
  ecbm_insert_channel(ch);
 } 
 spin_unlock(&comm_structs_spin); /* CRITICAL SECTION: Unlock structure, please */

 if(found)
 {
  kfree(ch);
  ch=found;
 }

 wake_up(&ch->acquire); /* Channel ready */

 /* Update channel fields */
 ch->n=asynchrony;
 ch->insert=0;
 ch->extract=0;
 ch->cap=(capability**)kmalloc(asynchrony*sizeof(capability*),GFP_KERNEL);

 /* Insert channel in the per-process channel table */
 if(!(chs=current->chs))
 { 
  chs=(struct channel_table*) kmalloc(sizeof(struct channel_table),GFP_KERNEL); 
  current->chs=chs;
  chs->free_id=0;
 }
 chs->addresses[chs->free_id]=ch;
 
 return chs->free_id++;
}

/*
 * Acquire a communication channel from kernel structures
 * @param magic_number_t magic - the progressive id of the channel
 * @return the address of the channel within the per-process channel table
 */
asmlinkage id_t ecbm_acquire_channel(magic_number_t magic)
{
 channel* ch;
 channel* found;
 struct wait_queue wait; 
 struct channel_table* chs;

 if(current->chs && current->chs->free_id>=ECBM_COMM_PP_MAX)
  return -EMFILE;

 ch=(channel*)kmalloc(sizeof(channel),GFP_KERNEL);

 /* Search and create channel */
 spin_lock(&comm_structs_spin); /* CRITICAL SECTION: Lock structure, please */
 if(!(found=ecbm_search_channel(magic)))
 {
  ch->magic=magic;
  ch->filled=0;
  init_waitqueue(&ch->acquire);
  ch->spin=SPIN_LOCK_UNLOCKED;
  init_waitqueue(&ch->full);
  init_waitqueue(&ch->empty);
  ecbm_insert_channel(ch);

  current->state = TASK_INTERRUPTIBLE;
  wait.task=current;
  add_wait_queue(&ch->acquire,&wait);
 }
 spin_unlock(&comm_structs_spin); /* CRITICAL SECTION: Unlock structure, please */

 if(!found) 
 {
  schedule();  /* Wait for channel ready */
  spin_lock(&ch->spin); /* CRITICAL SECTION: Lock structure, please */
  remove_wait_queue(&ch->acquire,&wait);
  spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */
 }
 else
 {
  kfree(ch);
  ch=found;
 }

 /* Insert channel in the per-process channel table */
 if(!(chs=current->chs))
 { 
  chs=(struct channel_table*) kmalloc(sizeof(struct channel_table),GFP_KERNEL); 
  current->chs=chs;
  chs->free_id=0;
 }
 chs->addresses[chs->free_id]=ch;

 return chs->free_id++;
}

/*
 * Search a channel into kernel structures
 * @param magic_number_t magic - the progressive id of the channel
 * @return the address of the channel, if any
 */
channel* ecbm_search_channel(magic_number_t magic)
{
 channel* ch;

 /* AVL? */
 if(avl_root)
 {
  ch=avl_root;
  while(ch && ch->magic!=magic)
  {
   if(ch->magic>magic) ch=ch->left;
   else ch=ch->right;
  }

  return ch;
 }

 /* ORDERED LIST */
 ch=list_head;
 while(ch && ch->magic!=magic)
  ch=ch->next;

 return ch;
}

/*
 * Insert a new channel into the kernel ordered list
 * @param channel* ch - the channel to be inserted
 * @return null
 */
void ecbm_insert_channel(channel* ch)
{
 channel* parent;
 channel* cur;

 /* First channel? */
 if(count==0)
 { 
  list_head=ch;
  ch->next=NULL;
  ++count;
  return;
 }

 /* AVL? */
 if(avl_root) ecbm_avl_insert(ch);

 /* ORDERED LIST */
 parent=NULL;
 cur=list_head;
 while(cur && cur->magic<ch->magic)
 {
  parent=cur;
  cur=cur->next;
 }

 if(parent) 
 {
  parent->next=ch;
  ch->next=cur;
 }
 else
 {
  ch->next=list_head;
  list_head=ch;
 }

 /* Do we have to create an AVL now? */
 ++count;
 if(count>ECBM_COMM_RES_MAX) ecbm_avl_build();
 return;
}

/*
 * Insert a new channel into the kernel avl tree
 * @param channel* ch - the channel to be inserted
 * @return null
 * @note THE AVL MUST ALREADY EXIST!!!
 */
void ecbm_avl_insert(channel* ch)
{
 channel* parent;
 channel* cur;

 /* Search position */
 parent=NULL;
 cur=avl_root;
 while(cur)
 {
  parent=cur;
  if(cur->magic>ch->magic) cur=cur->left;
  else cur=cur->right;
 }

 /* Insert new */
 if(parent->magic>ch->magic)
 {
  parent->left=ch;
  ++(parent->balance);
 }
 else
 {
  parent->right=ch;
  --(parent->balance);
 }

 ch->parent=parent;
 ch->left=NULL;
 ch->right=NULL;
 ch->balance=0;

 /* Update balancing factors */
 ecbm_avl_update(parent);
 return;
}

/*
 * Build up a new avl tree from the ordered list
 * @param null
 * @return null
 * @note THE ORDERED LIST MUST ALREADY EXIST!!!
 */
void ecbm_avl_build()
{
 channel* ch;
 
 /* First channel */
 avl_root=list_head;
 avl_root->parent=NULL;
 avl_root->left=NULL;
 avl_root->right=NULL;
 avl_root->balance=0;

 /* All the othres */
 ch=list_head->next;
 while(ch)
 {
  ecbm_avl_insert(ch);
  ch=ch->next;
 }
 return;
}

/*
 * Update balancing factors in the avl tree
 * @param channel* ch - the channel to start from
 * @return null
 */
void ecbm_avl_update(channel* ch)
{
 channel* parent;
 channel* cur;

 /* Iterate over tree, from ch to root */
 parent=ch->parent;
 cur=ch;
 while(parent)
 {
  if(parent->left==cur) ++(parent->balance);
  else --(parent->balance);
  
  /* Wrong balancing factor? */
  if(abs(parent->balance)>1) 
  {
   ecbm_avl_rebalance(parent);
   return;
  }
  else
  {
   cur=parent;
   parent=parent->parent;
  }
 }
 return;
}

/*
 * Rebalance the avl tree
 * @param channel* ch - the channel to start from
 * @return the new sub-root of the sub-tree 
 *         previously rooted in ch
 */
channel* ecbm_avl_rebalance(channel* ch)
{
 channel* new_ch;

 if(ch->balance>0)
 {
  /*
   * AVL Right Rotations
   */
  if(ch->left->balance>0)
  {
   /*
    * AVL Single Rotation:
    *
    *            K        <=== ch
    *          /   \
    *        I      Z[n]
    *      /   \ 
    *   X[n+1]  Y[n] 
    * 
    * After Rotation:
    *
    *            I        <=== new_ch
    *          /   \
    *       X[n+1]   K
    *              /   \
    *            Y[n]  Z[n]
    */
   channel* ex_y=ch->left->right;

   new_ch=ch->left;
   new_ch->parent=ch->parent;
   if(new_ch->parent)
   {
    if(new_ch->parent->left==ch) new_ch->parent->left=new_ch;
    else new_ch->parent->right=new_ch;
   }
   new_ch->right=ch;
   new_ch->balance=0;

   ch->parent=new_ch;
   ch->left=ex_y;
   ch->balance=0;

   if(ex_y)
    ex_y->parent=ch;
  }
  else
  {
   /*
    * AVL Double Rotation:
    *
    *            K        <=== ch
    *          /   \
    *        I     Z[n]
    *      /   \ 
    *    X[n]   J
    *         /   \
    *       A[?]  B[?]
    * 
    * After Rotation:
    *
    *             J        <=== new_ch
    *          /     \
    *        I         K
    *      /   \     /   \
    *    X[n] A[?] B[?] Z[n]
    */
   channel* ex_i=ch->left;
   channel* ex_a=ch->left->right->left;
   channel* ex_b=ch->left->right->right;
   int ex_balance=ch->left->right->balance;

   new_ch=ch->left->right;
   new_ch->parent=ch->parent;
   if(new_ch->parent)
   {
    if(new_ch->parent->left==ch) new_ch->parent->left=new_ch;
    else new_ch->parent->right=new_ch;
   }
   new_ch->left=ch->left;
   new_ch->right=ch;
   new_ch->balance=0;

   ch->parent=new_ch;
   ch->left=ex_b;
   ch->balance=(ex_balance>0) ? -1 : 0;

   ex_i->parent=new_ch;
   ex_i->right=ex_a;
   ex_i->balance=(ex_balance<0) ? +1 : 0;

   if(ex_a)
    ex_a->parent=ex_i;
   if(ex_b)
    ex_b->parent=ch;
  }
 }
 else
 {
  /*
   * AVL Left Rotations
   */
  if(ch->right->balance<0)
  {
   /*
    * AVL Single Rotation:
    *
    *            K        <=== ch
    *          /   \
    *        Z[n]    I
    *              /   \ 
    *           X[n]  Y[n+1] 
    * 
    * After Rotation:
    *
    *            I        <=== new_ch
    *          /   \
    *        K    Y[n+1]
    *      /   \
    *    Z[n]  X[n]
    */
   channel* ex_x=ch->right->left;

   new_ch=ch->right;
   new_ch->parent=ch->parent;
   if(new_ch->parent)
   {
    if(new_ch->parent->left==ch) new_ch->parent->left=new_ch;
    else new_ch->parent->right=new_ch;
   }
   new_ch->left=ch;
   new_ch->balance=0;

   ch->parent=new_ch;
   ch->right=ex_x;
   ch->balance=0;

   if(ex_x)
    ex_x->parent=ch;
  }
  else
  {
   /*
    * AVL Double Rotation:
    *
    *            K        <=== ch
    *          /   \
    *        Z[n]    I
    *              /   \ 
    *             J    X[n]   
    *           /   \
    *         A[?]  B[?]
    * 
    * After Rotation:
    *
    *             J        <=== new_ch
    *          /     \
    *        K         I
    *      /   \     /   \
    *    Z[n] A[?] B[?] X[n]
    */
   channel* ex_i=ch->right;
   channel* ex_a=ch->right->left->left;
   channel* ex_b=ch->right->left->right;
   int ex_balance=ch->right->left->balance;

   new_ch=ch->right->left;
   new_ch->parent=ch->parent;
   if(new_ch->parent)
   {
    if(new_ch->parent->left==ch) new_ch->parent->left=new_ch;
    else new_ch->parent->right=new_ch;
   }
   new_ch->left=ch;
   new_ch->right=ch->right;
   new_ch->balance=0;

   ch->parent=new_ch;
   ch->right=ex_a;
   ch->balance=(ex_balance<0) ? 1 : 0;

   ex_i->parent=new_ch;
   ex_i->left=ex_b;
   ex_i->balance=(ex_balance>0) ? -1 : 0;

   if(ex_a)
    ex_a->parent=ch;
   if(ex_b)
    ex_b->parent=ex_i;
  }
 }

 /* New root for AVL? */
 if(avl_root==ch) avl_root=new_ch;

 return new_ch;
}

/*
 * Send a message over a communication channel
 * @param id_t id - the address of the channel
 * @param void* buffer - the message to be sent
 * @param int size - the size of the message in bytes
 * @param prot_t rights - the access rights the receiver will be granted
 *                        (they will be anded with owner rights)
 * @return TRUE on success, set errno on error
 */
asmlinkage int ecbm_zc_send(id_t id, void* buffer, int size, prot_t rights)
{
 channel* ch;
 capability* cap;
 struct wait_queue wait;

 /* Some checks */
 if(id<0 || id>=current->chs->free_id) return -EBADF;
 ch=current->chs->addresses[id];

 /* Create capability */
 cap=ecbm_create_capability(buffer,size,rights);
 if(!cap)
  return -EINVAL;

 /* Insert */
 spin_lock(&ch->spin); /* CRITICAL SECTION: Lock structure, please */
 ch->cap[ch->insert]=cap;
 ch->insert=(ch->insert+1)%ch->n;
 ++ch->filled;

 /* Manage channel */
 if(ch->filled==1) wake_up(&ch->empty); /* Channel was empty */

 if(ch->filled!=ch->n)
 {
  spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */

  return TRUE;
 }

 current->state = TASK_INTERRUPTIBLE;
 wait.task=current;
 add_wait_queue(&ch->full,&wait);
 spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */

 schedule();  /* Wait for channel ready */
 spin_lock(&ch->spin); /* CRITICAL SECTION: Lock structure, please */
 remove_wait_queue(&ch->full,&wait);
 spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */

 return TRUE;
}

/*
 * Receive a message from a communication channel
 * @param id_t id - the address of the channel
 * @param prot_t rights - the access rights to be granted
 *                        (they will be anded with the rights the sender imposed 
 *                         on the structure being receved)
 * @return the message received
 */
asmlinkage void* ecbm_zc_receive(id_t id, prot_t rights)
{
 channel* ch;
 capability* cap;
 struct wait_queue wait;

 /* Some checks */
 if(id<0 || id>=current->chs->free_id) 
 {
  errno=-EBADF;
  return NULL;
 }
 ch=current->chs->addresses[id];

 spin_lock(&ch->spin); /* CRITICAL SECTION: Lock structure, please */
 /* Manage channel */
 if(!ch->filled) 
 {
  current->state = TASK_INTERRUPTIBLE;
  wait.task=current;
  add_wait_queue(&ch->empty,&wait);
  spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */

  schedule();  /* Wait for channel ready */
  spin_lock(&ch->spin); /* CRITICAL SECTION: Lock structure, please */
  remove_wait_queue(&ch->empty,&wait);
 }

 /* Extract */
 cap=ch->cap[ch->extract];
 ch->extract=(ch->extract+1)%ch->n;
 --ch->filled;

 /* Manage channel */
 if(ch->filled==ch->n-1) wake_up(&ch->full); /* Channel was full */
 spin_unlock(&ch->spin); /* CRITICAL SECTION: Unlock structure, please */

 return ecbm_acquire_capability(cap,rights);
} 

/*
 * Send a message over a communication channel
 * @param id_t id - the address of the channel
 * @param void* buffer - the message to be sent
 * @param int size - the size of the message in bytes
 * @return TRUE on success, set errno on error
 */
asmlinkage int ecbm_oc_send(id_t id, void* buffer, int size)
{
 return -ENOSYS;
}

/*
 * Receive a message from a communication channel
 * @param id_t id - the address of the channel
 * @return the message received
 */
asmlinkage void* ecbm_oc_receive(id_t id)
{  
 errno=-ENOSYS;
 return NULL;
} 









